#include<stdio.h>
#include<math.h>
#include<complex.h>

int main(){

  double theta, theta_rad;
  /* 倍精度複素数型の2x2の２次元配列を定義 */
  double _Complex sigmax[2][2], sigmay[2][2], sigmaz[2][2], amatx[2][2], amaty[2][2], amatz[2][2], cmatx[2][2], cmaty[2][2], cmatz[2][2], unit[2][2];
  int i,j;

  /* 倍精度複素数型の配列を指すポインタ */
  double _Complex *amatxp, *amatyp, *amatzp;

  /* 行列指数関数exp(A)を計算する関数、第一引数は倍精度複素数型のポインタ */
  int expAtaylor(double _Complex *, int, int);

  /* sigmax, sigmay, sigmazと単位行列unitに値を代入。Iはcomplex.hで定義された虚数単位 */
  sigmax[0][0] = 0.0;   sigmax[0][1] = 1.0;   sigmax[1][0] = 1.0;   sigmax[1][1] = 0.0;
  sigmay[0][0] = 0.0;   sigmay[0][1] =  -I;   sigmay[1][0] =   I;   sigmay[1][1] = 0.0;
  sigmaz[0][0] = 1.0;   sigmaz[0][1] = 0.0;   sigmaz[1][0] = 0.0;   sigmaz[1][1] =-1.0;
  unit  [0][0] = 1.0;   unit  [0][1] = 0.0;   unit  [1][0] = 0.0;   unit  [1][1] = 1.0; 

  /* 角度は-180度から30度ごとに計算 */
  for(theta =-180.0; theta<181.0; theta+=30){

    /* 角度をラジアンに変換。M_PIはmath.hで定義される円周率 */
    theta_rad = theta * M_PI/180.0;
    printf("theta = %f\n", theta);

    /* amat(x,y,z)にi*theta*sigma(x,y,z)を代入 */
    for(i=0;i<2;i++){
      for(j=0;j<2;j++){
	amatx[i][j] = I*theta_rad*sigmax[i][j];
	amaty[i][j] = I*theta_rad*sigmay[i][j];
	amatz[i][j] = I*theta_rad*sigmaz[i][j];
      }
    }

    /* amat(x,y,z)pを二次元配列のamat(x,y,z)を指すようにセット */ 
    amatxp = (double _Complex *) amatx;
    amatyp = (double _Complex *) amaty;
    amatzp = (double _Complex *) amatz;

    /* 2x2の行列指数関数をテーラー展開で30次まで計算。行列はamat(x,y,z)pでポインタで渡し、結果はamat(x,y,z)に上書き */
    expAtaylor(amatxp,2,30);
    expAtaylor(amatyp,2,30);
    expAtaylor(amatzp,2,30);

    /* 行列指数関数の結果の出力 */
    printf("exp(i*theta*sigma)\n");
    for(i=0;i<2;i++){
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(amatx[i][j]), cimag(amatx[i][j]));
      }
      printf("        ");
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(amaty[i][j]), cimag(amaty[i][j]));
      }
      printf("        ");
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(amatz[i][j]), cimag(amatz[i][j]));
      }
      printf("\n");
    }

    /* I cos(theta) + i sigma(x,y,z) sin(theta)をcmat(x,y,z)に代入 */
    for(i=0;i<2;i++){
      for(j=0;j<2;j++){    
	cmatx[i][j] = unit[i][j] * cos(theta_rad) + I * sigmax[i][j]* sin(theta_rad);
	cmaty[i][j] = unit[i][j] * cos(theta_rad) + I * sigmay[i][j]* sin(theta_rad);
	cmatz[i][j] = unit[i][j] * cos(theta_rad) + I * sigmaz[i][j]* sin(theta_rad);
      }
    }

    /* cmat(x,y,z)の出力 */
    printf("I*cos(theta) + i*sigma*sin(theta)\n");
    for(i=0;i<2;i++){
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(cmatx[i][j]), cimag(cmatx[i][j]));
      }
      printf("        ");
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(cmaty[i][j]), cimag(cmaty[i][j]));
      }
      printf("        ");
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(cmatz[i][j]), cimag(cmatz[i][j]));
      }
      printf("\n");
    }
    
  }

  return 0;
  
}


int expAtaylor(double _Complex *amatp, int dim, int n){
  /* dim*dim次元の行列のexp(A)をn次までのテーラー展開で計算する関数 
     inputの行列Aはポインタ*amatpで指定される一次元配列
     outputの行列exp(A)はポインタamatpで指定される一次元配列が上書き代入される */

  int i,j,k;
  double _Complex amat0[dim*dim], amatk[dim*dim];
  double _Complex *amatkp;
  
  /* 行列Aのn乗を計算する関数 */
  int powmat(double _Complex *, int, int);
  /* 階乗 */
  double factorial(int);

  /* inputの行列Aを配列amat0にコピー。amatpの値を初期化 */
  for(i=0;i<dim*dim;i++){
    amat0[i]=amatp[i];
    amatp[i] = 0.0;
  }

  /* テーラー展開の次数kのループ */
  for(k=0;k<=n;k++){

    /* 行列Aを配列amatkにコピー */
    for(i=0;i<dim*dim;i++){
      amatk[i]=amat0[i];
    }

    /* ポインタamatkpをamatk(行列A)を指すようにセット */
    amatkp = (double _Complex *) amatk;
    /* amatk(行列A)のk乗を計算、計算結果(A^k)をamatkに上書き */
    powmat(amatkp,dim,k);

    /* テーラー展開の和A^k/k!を計算 */
    for(i=0;i<dim*dim;i++){
      amatp[i] += amatkp[i]/factorial(k);
    }
    
  }
  
  return (0);
  
}

/* dim*dim行列amatkのk乗を計算 */
int powmat(double _Complex *amatk, int dim, int k){

  double _Complex amat[dim*dim], amat0[dim*dim];
  double _Complex *amatk2;
  int i,j,l;

  /* 行列amatkの内容をamatとamat0にコピー */
  for(i=0;i<dim*dim;i++){
    amat[i]   = amatk[i];
    amat0[i]  = amatk[i];
  }
  
  /* k<0の場合はエラー */
  if(k < 0){
    printf("k must be zero or positive\n");
    return -1;
  } else if(k==0){
    /* k=0の場合はamatkに単位行列を代入 */
   
    for(i=0;i<dim;i++){
      for(j=0;j<dim;j++){
	amatk[i*dim+j] = 0.0;
	if(i==j) amatk[i*dim+j] = 1.0;
	/* 一次元化された行列のインデックスはi*dim+jとなる */
      }
    }
    return 0;
    
  } else if(k==1){
    /* k=1の場合はamatkをそのまま返すので値を変更せずに終了 */
    return 0;
  } else {
    /* k>1の場合は再帰的に計算 */

    /* amatk2ポインタでamat配列を指すようにセット */
    amatk2=(double _Complex *) amat;
    /* powmatを再帰的に呼び出しamatのk-1乗を計算 */
    powmat(amatk2,dim,k-1);

    /* amatkの値を初期化 */
    for(i=0;i<dim*dim;i++){
      amatk[i]  = 0.0;
    }

    /* amatkはamatのk-1乗(amatk2)とamat(amat0)の行列積 */
    for(i=0;i<dim;i++){
      for(j=0;j<dim;j++){
	for(l=0;l<dim;l++){
	  amatk[i*dim+j] += amatk2[i*dim+l] * amat0[l*dim+j];
	  /* それぞれ[i][j], [i][l], [l][j]成分に対応 */
	}
      }
    }
    
  }
    
  return 0;
  
}
  


/* 階乗の計算 */ 
double factorial(int n){
    
  int k;
  double fact;

  fact = 0.0;
  if(n<0) return (fact);

  fact = 1.0;
  if(n==0) return(fact);

  for(k=1;k<=n;k++){

    fact *= (double) k;

  }
    
  return(fact);

}
